from unittest.mock import MagicMock

import pytest
from sqlalchemy import (
    desc,
    select,
)
from sqlalchemy.ext.asyncio import AsyncSession

from fastapi_toolkit.crud.base import CRUDBase
from tests.mixer import AsyncMixer
from tests.models import Book
from tests.schemas import (
    BookSchema,
    BookSchemaCreate,
)


class TestCRUDGet:
    async def test_one__ok(self, mixer: AsyncMixer):
        test_book = await mixer.blend(Book)
        book = await CRUDBase(Book).get(Book.title == test_book.title)
        assert BookSchema.from_orm(book) == BookSchema.from_orm(test_book)

    async def test_no_one__ok(self, mixer: AsyncMixer):
        await mixer.blend(Book, title='some_title')
        book = await CRUDBase(Book).get(Book.title == 'wrong_title')
        assert book is None


class TestCRUDCount:
    async def test_count__ok(self, mixer: AsyncMixer):
        test_count = 3
        await mixer.cycle(test_count).blend(Book)
        count = await CRUDBase(Book).count()
        assert count == test_count

    async def test_count_with_filter__ok(self, mixer: AsyncMixer):
        test_count = 3
        await mixer.cycle(test_count).blend(
            Book,
            title=mixer.sequence('Title_{0}'),
            page_amount=mixer.sequence()
        )
        count = await CRUDBase(Book).count((Book.title == 'Title_1', ))
        assert count == 1


class TestCRUDGetModelId:
    def test__ok(self):
        id_mock = MagicMock()
        model_mock = MagicMock(id=id_mock)
        crud = CRUDBase(model_mock)
        assert crud.get_model_id() is id_mock


class TestCRUDList:
    async def test__ok(self, mixer: AsyncMixer):
        test_count = 3
        test_books = await mixer.cycle(test_count).blend(Book)
        books, count = await CRUDBase(Book).list()

        assert count == test_count
        for book, test_book in zip(
            sorted(books, key=lambda x: x.id), test_books
        ):
            assert BookSchema.from_orm(book) == BookSchema.from_orm(test_book)

    async def test_empty_list__ok(self, mixer: AsyncMixer):
        books, count = await CRUDBase(Book).list()
        assert list(books) == []
        assert count == 0

    async def test_offset__ok(self, mixer: AsyncMixer):
        test_count = 3
        await mixer.cycle(test_count).blend(Book)
        books, count = await CRUDBase(Book).list(offset=1)
        assert len(list(books)) == 2
        assert count == test_count

    async def test_limit__ok(self, mixer: AsyncMixer):
        test_count = 3
        await mixer.cycle(test_count).blend(Book)
        books, count = await CRUDBase(Book).list(limit=2)
        assert len(list(books)) == 2
        assert count == test_count

    async def test_filter_query__ok(self, mixer: AsyncMixer):
        test_count = 3
        test_books = await mixer.cycle(test_count).blend(
            Book,
            title=mixer.sequence('Title_{0}'),
        )
        books, count = await CRUDBase(Book).list(
            filter_query=(Book.title == 'Title_1', )
        )
        assert count == 1
        assert BookSchema.from_orm(list(books)[0]) == BookSchema.from_orm(
            test_books[1]
        )

    async def test_estimated_count_false__ok(self, mixer: AsyncMixer):
        await mixer.cycle(3).blend(Book)
        _, count = await CRUDBase(Book).list(estimate_count=False)

        assert count is None

    async def test_order_by__ok(self, mixer: AsyncMixer):
        test_count = 3
        test_books = await mixer.cycle(test_count).blend(
            Book,
            page_amount=mixer.sequence(7, 3, 4),
        )
        books, count = await CRUDBase(Book).list(order_by=(Book.page_amount, ))

        assert count == test_count
        for book, test_book in zip(
            books, sorted(test_books, key=lambda x: x.page_amount)
        ):
            assert BookSchema.from_orm(book) == BookSchema.from_orm(test_book)

    async def test_multiple_params__ok(self, mixer: AsyncMixer):
        test_count = 6
        test_books = await mixer.cycle(test_count).blend(
            Book,
            page_amount=mixer.sequence(7, 3, 4, 5, 9, 10),
        )
        books, count = await CRUDBase(Book).list(
            filter_query=(Book.page_amount > 4, ),
            limit=1,
            offset=1,
            order_by=(desc(Book.page_amount), )
        )
        books_list = list(books)
        assert count == 4
        assert len(books_list) == 1
        assert BookSchema.from_orm(books_list[0]) == BookSchema.from_orm(
            test_books[-2]
        )


class TestCRUDCreate:
    @pytest.mark.parametrize(
        ('args', 'kwargs'),
        [
            ((BookSchemaCreate(title='test_title', page_amount=5), ), {}, ),
            ((BookSchemaCreate(title='test_title'), {'page_amount': 5}), {}, ),
            ((), {'title': 'test_title', 'page_amount': 5}, )
        ]
    )
    async def test__ok(
            self, args, kwargs, session: AsyncSession
    ):
        result = await CRUDBase(Book).create(*args, **kwargs)
        book_from_db = (await session.execute(select(Book))).scalars().first()
        test_result = {
            'title': 'test_title', 'page_amount': 5, 'id': book_from_db.id
        }
        assert BookSchema.from_orm(result) == BookSchema.from_orm(book_from_db)
        assert BookSchema.from_orm(result) == BookSchema(**test_result)


class TestCRUDUpdate:
    async def test_db_obj__ok(self, mixer: AsyncMixer):
        init_book = await mixer.blend(Book, title='title', page_amount=5)
        test_title = 'new_title'
        init_book.title = test_title

        new_book = await CRUDBase(Book).update(
            init_book, session=mixer.session
        )
        book_from_db = (
            await mixer.session.execute(select(Book))
        ).scalars().first()

        assert BookSchema.from_orm(new_book) == BookSchema.from_orm(
            book_from_db
        )
        assert BookSchema.from_orm(new_book) == BookSchema(
            id=book_from_db.id,
            title=test_title,
            page_amount=5
        )

    async def test_obj_in__ok(self, mixer: AsyncMixer):
        init_book = await mixer.blend(Book, title='title', page_amount=5)
        test_title = 'new_title'
        obj_in = BookSchemaCreate(title=test_title)
        new_book = await CRUDBase(Book).update(
            init_book, obj_in, session=mixer.session
        )
        book_from_db = (
            await mixer.session.execute(select(Book))
        ).scalars().first()
        assert BookSchema.from_orm(new_book) == BookSchema.from_orm(
            book_from_db
        )
        assert BookSchema.from_orm(new_book) == BookSchema(
            id=book_from_db.id,
            title=test_title,
            page_amount=5
        )

    async def test_data__ok(self, mixer: AsyncMixer):
        init_book = await mixer.blend(Book, title='title', page_amount=5)
        test_title = 'new_title'
        data = {'title': test_title}
        new_book = await CRUDBase(Book).update(
            init_book, session=mixer.session, **data
        )
        book_from_db = (
            await mixer.session.execute(select(Book))
        ).scalars().first()
        assert BookSchema.from_orm(new_book) == BookSchema.from_orm(
            book_from_db
        )
        assert BookSchema.from_orm(new_book) == BookSchema(
            id=book_from_db.id,
            title=test_title,
            page_amount=5
        )


class TestCRUDDelete:
    async def test__ok(self, mixer: AsyncMixer):
        book_1, book_2 = await mixer.cycle(2).blend(Book)
        await CRUDBase(Book).delete(Book.id == book_2.id)
        books_from_db = (await mixer.session.execute(select(Book))).scalars()
        books_from_db_list = list(books_from_db)
        assert len(books_from_db_list) == 1
        assert BookSchema.from_orm(book_1) == BookSchema.from_orm(
            books_from_db_list[0]
        )


class TestCRUDGetOrCreate:
    async def test_create__ok(self, session: AsyncSession):
        test_title = 'test_title'
        defaults = {
            'title': test_title,
            'page_amount': 5,
        }
        result, created = await CRUDBase(Book).get_or_create(
            Book.title == test_title, **defaults
        )
        book_from_db = (await session.execute(select(Book))).scalars().first()
        test_result = {
            'id': book_from_db.id,
            **defaults
        }
        assert BookSchema.from_orm(result) == BookSchema.from_orm(book_from_db)
        assert BookSchema.from_orm(result) == BookSchema(**test_result)
        assert created is True

    async def test_get__ok(self, mixer: AsyncMixer):
        test_title = 'test_title'
        defaults = {
            'title': test_title,
            'page_amount': 5,
        }
        await mixer.blend(Book, **defaults)
        result, created = await CRUDBase(Book).get_or_create(
            Book.title == test_title, **defaults
        )
        book_from_db = (
            await mixer.session.execute(select(Book))
        ).scalars().first()
        test_result = {
            'id': book_from_db.id,
            **defaults
        }
        assert BookSchema.from_orm(result) == BookSchema.from_orm(book_from_db)
        assert BookSchema.from_orm(result) == BookSchema(**test_result)
        assert created is False



from pydantic import BaseModel


class ApplicationInfoGit(BaseModel):
    hash: str = None
    branch: str = None


class ApplicationInfo(BaseModel):
    project: str = None
    version: str = None
    datetime: str
    environment: str
    git: ApplicationInfoGit



import os

import pytest
from sqlalchemy.ext.asyncio import AsyncSession

from tests.mixer import AsyncMixer

os.environ.setdefault('SETTINGS_MODULE', 'tests.conf')  # noqa

from fastapi_toolkit.db import (
    BaseModel,
    create_engine,
    create_session,
)
from tests.models import *  # noqa


@pytest.fixture
async def db():
    engine = create_engine()

    async with engine.begin() as connection:
        await connection.run_sync(BaseModel.metadata.create_all)

    yield

    async with engine.begin() as connection:
        await connection.run_sync(BaseModel.metadata.drop_all)

    await engine.dispose()


@pytest.fixture
async def session(db) -> AsyncSession:
    async with create_session() as session:
        yield session


@pytest.fixture
async def mixer(session) -> AsyncMixer:
    yield AsyncMixer(session)



import asyncio
import logging
from asyncio import ensure_future
from functools import wraps
from traceback import format_exception
from typing import (
    Any,
    Callable,
    Coroutine,
    Optional,
    Union,
)

from starlette.concurrency import run_in_threadpool

__all__ = (
    'repeat_every',
)

NoArgsNoReturnFuncT = Callable[[], None]
NoArgsNoReturnAsyncFuncT = Callable[[], Coroutine[Any, Any, None]]
NoArgsNoReturnDecorator = Callable[[Union[NoArgsNoReturnFuncT, NoArgsNoReturnAsyncFuncT]], NoArgsNoReturnAsyncFuncT]  # noqa


def repeat_every(
    *,
    seconds: float,
    wait_first: bool = False,
    logger: Optional[logging.Logger] = None,
    raise_exceptions: bool = False,
    max_repetitions: Optional[int] = None,
) -> NoArgsNoReturnDecorator:

    def decorator(func: Union[NoArgsNoReturnAsyncFuncT, NoArgsNoReturnFuncT]) -> NoArgsNoReturnAsyncFuncT:  # noqa
        is_coroutine = asyncio.iscoroutinefunction(func)

        @wraps(func)
        async def wrapped() -> None:
            repetitions = 0

            async def loop() -> None:
                nonlocal repetitions
                if wait_first:
                    await asyncio.sleep(seconds)
                while max_repetitions is None or repetitions < max_repetitions:
                    try:
                        if is_coroutine:
                            await func()  # type: ignore
                        else:
                            await run_in_threadpool(func)
                        repetitions += 1
                    except Exception as exc:
                        if logger is not None:
                            formatted_exception = ''.join(format_exception(
                                type(exc), exc, exc.__traceback__
                            ))
                            logger.error(formatted_exception)
                        if raise_exceptions:
                            raise exc
                    await asyncio.sleep(seconds)

            ensure_future(loop())

        return wrapped

    return decorator



import os
import subprocess

from fastapi_toolkit.conf import settings


def dbshell():
    os.environ['PGPASSWORD'] = settings.database_dsn.password
    subprocess.call(
        [
            'psql',
            '-h', settings.database_dsn.host,
            '-p', str(settings.database_dsn.port),
            '-U', settings.database_dsn.user,
            '-d', settings.database_dsn.path[1:],  # slice '/' from path
        ]
    )



from typing import Union

from pydantic import (
    AnyUrl,
    BaseSettings,
    Field,
)

__all__ = (
    'GlobalSettings',
)

PortType = Union[str, int]


class GlobalSettings(BaseSettings):
    environment: str = 'local'
    project: str = None
    debug: bool = False
    version: str = None
    application: str = None
    alembic_config: str = None
    alembic_directory: str = 'alembic'
    use_https: bool = True
    commands_roots: list[str] = Field(default_factory=list)

    database_dsn: AnyUrl = None

    log_dir: str = None
    log_level: str = 'DEBUG'

    app_schema: str = 'http'
    app_host: str = None
    app_port: PortType = None

    git_branch: str = 'local'
    git_hash: str = 'local'

    tag_groups: list[dict] = Field(default_factory=list)

    sentry_enabled: bool = True
    sentry_dsn: str = None

    new_relic_enabled: bool = True

    redis_dsn: AnyUrl = None
    redis_max_connections: int = 10

    @property
    def logging(self) -> dict:
        return {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'consoleFormatter': {
                    '()': 'uvicorn.logging.DefaultFormatter',
                    'fmt': '[%(process)d] [%(asctime)s] [%(levelname)s] %(name)s -> %(message)s'  # noqa
                },
                'access': {
                    '()': 'uvicorn.logging.AccessFormatter',
                    'fmt': '[%(asctime)s] [%(levelname)s] %(name)s -> "%(request_line)s" %(status_code)s'  # noqa
                }
            },
            'handlers': {
                'console': {
                    'level': 'DEBUG',
                    'class': 'logging.StreamHandler',
                    'stream': 'ext://sys.stdout',
                    'formatter': 'consoleFormatter'
                },
                'access': {
                    'formatter': 'access',
                    'class': 'logging.StreamHandler',
                    'stream': 'ext://sys.stdout'
                }
            },
            'loggers': {
                self.project: {
                    'handlers': ['console'],
                    'level': self.log_level
                },
                'uvicorn.error': {
                    'level': 'INFO'
                },
                'uvicorn.access': {
                    'handlers': ['access'],
                    'level': 'INFO',
                    'propagate': False
                }
            }
        }

    def update(self, attrs):
        for key, value in attrs.items():
            setattr(self, key, value)

    class Config:
        env_prefix = 'app_'
        env_file = '.env'
        env_nested_delimiter = '__'



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Awaitable,
    Optional,
    Type,
    TypeVar,
)

from mixer.backend.sqlalchemy import Mixer
from mixer.main import LOGGER
from sqlalchemy.ext.asyncio import AsyncSession

if TYPE_CHECKING:
    from fastapi_toolkit.db import BaseModel


TargetT = TypeVar('TargetT')


class AsyncProxyMixer:
    def __init__(
            self,
            mixer: AsyncMixer,
            count: int = 5,
            guards: Optional[tuple[list, dict]] = None
    ):
        self.count = count
        self.mixer = mixer
        self.guards = guards

    async def blend(
            self,
            scheme: Type[BaseModel],
            **values
    ) -> list[Awaitable]:
        if self.guards:
            return await self.mixer._guard(scheme, self.guards, **values)

        return [
            await self.mixer.blend(scheme, **values) for _ in range(self.count)
        ]


class AsyncMixer(Mixer):
    @property
    def session(self) -> Optional[AsyncSession]:
        return self.params.get('session')

    @property
    def commit(self) -> bool:
        return self.params.get('commit') or False

    async def _postprocess(self, target: TargetT) -> TargetT:
        if self.commit:
            session = self.session
            if not session:
                LOGGER.warning(
                    "'commit' set true but session not initialized."
                )
            else:
                session.add(target)
                await session.commit()

        return target

    def postprocess(self, target: TargetT) -> TargetT:
        return self._postprocess(target)

    def cycle(self, count: int = 5) -> AsyncProxyMixer:
        return AsyncProxyMixer(self, count)



from typing import TypeVar

from pydantic import BaseModel
from sqlalchemy import (
    delete,
    func,
    select,
)
from sqlalchemy.dialects.postgresql import insert

from fastapi_toolkit.db import create_session
from fastapi_toolkit.db.base_class import BaseModel as Model

ModelType = TypeVar('ModelType', bound=Model)
CreateSchemaType = TypeVar('CreateSchemaType', bound=BaseModel)
UpdateSchemaType = TypeVar('UpdateSchemaType', bound=BaseModel)


class CRUDBase:
    def __init__(self, model):
        self.model = model

    def get_model_id(self):
        return self.model.id

    @staticmethod
    def _get_filtered_queryset(exp, filter_query):
        for query in filter_query:
            exp = exp.filter(query)
        return exp

    @staticmethod
    def _get_paginated_queryset(exp, offset, limit):
        if limit is not None:
            exp = exp.limit(limit)
        if offset is not None:
            exp = exp.offset(offset)
        return exp

    @staticmethod
    def _get_sorted_queryset(exp, order_by):
        if order_by:
            exp = exp.order_by(*order_by)
        return exp

    def join(self, query):
        return query

    async def count(self, filter_query=tuple(), session=None):
        queryset = self.join(select(func.count(self.get_model_id())))
        queryset = self._get_filtered_queryset(queryset, filter_query)
        async with create_session(session) as session:
            count = (await session.execute(queryset)).scalar()
        return count

    async def get(
            self,
            condition,
            session=None
    ) -> ModelType:
        async with create_session(session) as session:
            return (await session.execute(
                select(self.model).filter(condition)
            )).scalars().first()

    async def list(
            self,
            offset: int = 0,
            limit: int = 100,
            order_by: tuple = tuple(),
            filter_query: tuple = tuple(),
            estimate_count: bool = True,
            session=None
    ) -> tuple[list[ModelType], int]:

        queryset = self.join(select(self.model))
        queryset = self._get_filtered_queryset(queryset, filter_query)
        queryset = self._get_sorted_queryset(queryset, order_by)
        queryset = self._get_paginated_queryset(queryset, offset, limit)

        count = None
        async with create_session(session) as session:
            items = (await session.execute(queryset)).scalars()
            if estimate_count:
                count = await self.count(filter_query, session=session)
        return items, count

    async def create(
            self,
            obj_in: CreateSchemaType = None,
            additional_data: dict = None,
            session=None,
            commit=True,
            **data,
    ) -> ModelType:
        if obj_in:
            data = obj_in.dict()
        if additional_data:
            data.update(additional_data)
        item = self.model(
            **data
        )
        async with create_session(session) as session:
            session.add(item)
            if commit:
                await session.commit()
                await session.refresh(item)
        return item

    async def update(
            self,
            db_obj: ModelType,
            obj_in: UpdateSchemaType = None,
            session=None,
            commit=True,
            **update_data
    ) -> ModelType:
        if obj_in:
            update_data = obj_in.dict(exclude_unset=True)
        async with create_session(session) as session:
            await session.merge(db_obj)
            for k, v in update_data.items():
                setattr(db_obj, k, v)
            if commit:
                await session.commit()
                await session.refresh(db_obj)
        return db_obj

    async def delete(self, condition, session=None, commit=True):
        async with create_session(session) as session:
            await session.execute(
                delete(self.model).filter(condition)
            )
            if commit:
                await session.commit()

    async def get_or_create(
            self,
            condition,
            session=None,
            commit=True,
            **defaults
    ):
        created = False
        async with create_session(session) as session:
            obj = await self.get(condition, session)
            if obj is None:
                orm_stmt = select(self.model).from_statement(
                    insert(self.model).values(
                        **defaults
                    ).on_conflict_do_nothing().returning(self.model)
                ).execution_options(populate_existing=True)
                obj = (await session.execute(orm_stmt)).scalar()
                if commit:
                    await session.commit()
                # sometimes previous statement doesn't return inserted object
                if not obj:
                    obj = await self.get(condition, session)
                created = True
        return obj, created



from pydantic import AnyUrl

from fastapi_toolkit.conf.global_settings import GlobalSettings


class Settings(GlobalSettings):
    project: str = '{{cookiecutter.project_name}}'
    alembic_directory: str = 'sources/alembic'
    alembic_config: str = 'sources/alembic/alembic.ini'
    application = 'sources.main:app'
    use_https = False
    database_dsn: AnyUrl = 'postgresql+asyncpg://{{cookiecutter.database_user}}:{{cookiecutter.database_password}}@{{cookiecutter.database_host}}:{{cookiecutter.database_port}}/{{cookiecutter.database_name}}'  # noqa



from fastapi import Query
from sqlalchemy import (
    asc,
    desc,
)

__all__ = ('order_by_fields', )


def order_by_fields(available_fields: dict, default=None):
    _fields = list(available_fields.keys())
    _fields.extend([f'-{_field}' for _field in _fields])

    def order_dependence(
        order_by_params: list = Query(
            [],
            alias='order_by[]',
            description=f'Available fields: {", ".join(_fields)}'
        )
    ) -> tuple:
        result_fields = []
        for field in order_by_params:
            direction = desc if field.startswith('-') else asc
            field = field.lstrip('-')
            if field in available_fields:
                result_fields.append(direction(available_fields[field]))

        _order_by = tuple(result_fields)
        if not _order_by and default is not None:
            _order_by = (default,) if not isinstance(default, tuple) else default  # noqa
        return _order_by
    return order_dependence



# Standard Library
import logging

# Third Party Library
import newrelic.agent
import sentry_sdk
from fastapi import FastAPI
from newrelic.agent import ASGIApplicationWrapper
from newrelic.api.transaction import current_transaction
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from sentry_sdk.integrations.httpx import HttpxIntegration
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
from starlette.types import ASGIApp

from fastapi_toolkit.conf import settings


class NewRelicMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope, receive, send):
        transaction = current_transaction()
        if transaction:
            _method = scope.get('method', '')
            _path = scope.get('path', '/')
            transaction_name = f'{_method}: {_path}'
            newrelic.agent.set_transaction_name(transaction_name, 'Uri')
            transaction._name_priority = None
        return await self.app(scope, receive, send)


def init_newrelic(app: FastAPI, newrelic_ini_path):
    if settings.new_relic_enabled:
        newrelic.agent.initialize(
            newrelic_ini_path,
            environment=settings.environment
        )

        app.add_middleware(NewRelicMiddleware)
        app = ASGIApplicationWrapper(app)

    return app


def init_sentry(app):
    if settings.sentry_enabled and settings.sentry_dsn:
        sentry_sdk.init(
            dsn=settings.sentry_dsn,
            environment=settings.environment,
            integrations=[
                LoggingIntegration(
                    level=logging.INFO,
                    event_level=logging.WARNING
                ),
                SqlalchemyIntegration(),
                HttpxIntegration()
            ]
        )
        app.add_middleware(SentryAsgiMiddleware)
    return app



# Standard Library
import os  # isort:skip
os.environ.setdefault('SETTINGS_MODULE', 'sources.conf')  # noqa

from sources.api import api_router as v1_api_router

# Third Party Library
from fastapi_toolkit.application import app

app.include_router(v1_api_router, prefix='/v1')


def get_app():
    return app



from fastapi import (
    HTTPException,
    status,
)
from sqlalchemy.exc import IntegrityError as DBIntegrityError


class IntegrityError:
    def __init__(self, raise_exc: bool = True):
        self._raise = raise_exc

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._raise and exc_val and isinstance(exc_val, DBIntegrityError):
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail='\n'.join(exc_val.orig.args)
            )
        return not self._raise



from distutils.util import strtobool

from sqlalchemy import func


def gte(x, y):
    return x >= y


def lte(x, y):
    return x <= y


def gt(x, y):
    return x > y


def lt(x, y):
    return x < y


def eq(x, y):
    return x == y


def like(x, y):
    return x.like('%{}%'.format(y))


def ilike(x, y):
    return x.ilike('%{}%'.format(y))


def in_(x, y):
    return x.in_(y)


def contained_by(x, y):
    return x.contained_by(y)


def contains(x, y):
    return x.contains(list(y))


def contains_like(x, y):
    return func.array_to_string(x, ',').like('%{}%'.format(y))


def empty_list(x, y):
    return x == [] if strtobool(y) else x != []


def is_null(x, y):
    return x.is_(None) if y else x.isnot(None)


FILTERS_LIST = (
    gte, lte, gt, lt, eq, like, in_,
    contained_by, contains, contains_like, empty_list, is_null
)

LIST_OPERATORS = {in_, contained_by, contains}
OPERATORS_TYPES = {
    is_null: bool
}



from fastapi_toolkit.dependencies import BasePagination

__all__ = (
    'api_response',
)


def api_response(data, count=None, pagination: BasePagination = None) -> dict:
    if isinstance(data, (list, tuple)):
        return dict(
            meta=pagination.schema(count=count, **pagination.params),
            result=data
        )
    return data



from dataclasses import dataclass
from typing import (
    Any,
    List,
)
from unittest.mock import (
    MagicMock,
    patch,
)

import pytest

from fastapi_toolkit.dependencies import filter_by_fields
from fastapi_toolkit.filters import filter_
from fastapi_toolkit.filters.func import (
    contained_by,
    contains,
    eq,
    gt,
    in_,
    is_null,
    lte,
)
from tests.models import Book


@dataclass
class QueryMock:
    default: Any
    alias: str


@patch('fastapi_toolkit.dependencies.filter.Query', QueryMock)
class TestFilterResult:
    @pytest.mark.parametrize(
        ('test_fields', 'result_fields'),
        (
            (
                {
                    'title': filter_(Book.title, (eq,)),
                    'page_amount': filter_(Book.page_amount, (gt, lte), str),
                },
                {
                    'title__eq': QueryMock(None, alias='title__eq'),
                    'page_amount__gt': QueryMock(
                        None, alias='page_amount__gt'
                    ),
                    'page_amount__lte': QueryMock(
                        None, alias='page_amount__lte'
                    )
                }
            ),
            (
                {
                    'title': filter_(Book.title, (in_, contains)),
                    'page_amount': filter_(
                        Book.page_amount, (is_null, contained_by), str
                    ),
                },
                {
                    'title__in': QueryMock(None, alias='title__in[]'),
                    'title__contains': QueryMock(
                        None, alias='title__contains[]'
                    ),
                    'page_amount__is_null': QueryMock(
                        None, alias='page_amount__is_null'
                    ),
                    'page_amount__contained_by': QueryMock(
                        None, alias='page_amount__contained_by[]'
                    )
                }
            ),
        )
    )
    def test_dataclass__ok(
            self, test_fields: dict, result_fields: dict
    ):
        result = filter_by_fields(test_fields)
        assert result.__name__ == 'FilterQueryParams'
        for field_name in result.__dataclass_fields__:
            assert getattr(result, field_name) == result_fields[field_name]

    @pytest.mark.parametrize(
        ('test_fields', 'result_annotations'),
        (
            (
                {
                    'title': filter_(Book.title, (eq,)),
                    'page_amount': filter_(Book.page_amount, (gt, lte), str),
                },
                {
                    'title__eq': str,
                    'page_amount__gt': int,
                    'page_amount__lte': int,
                }
            ),
            (
                {
                    'title': filter_(Book.title, (in_, contains)),
                    'page_amount': filter_(
                        Book.page_amount, (is_null, contained_by), str
                    ),
                },
                {
                    'title__in': List[str],
                    'title__contains': List[str],
                    'page_amount__is_null': bool,
                    'page_amount__contained_by': List[int]
                },
            ),
            (
                {
                    'title': filter_(
                        Book.title, (eq, contains), query_param_type=int
                    ),
                    'page_amount': filter_(Book.title, (gt, ), int, str),
                },
                {
                    'title__eq': int,
                    'title__contains': List[int],
                    'page_amount__gt': str,
                }
            )
        )
    )
    def test_dataclass__annotations_ok(
            self, test_fields: dict, result_annotations: dict
    ):
        result = filter_by_fields(test_fields)
        assert result.__name__ == 'FilterQueryParams'
        for field_name, annotation in result.__annotations__.items():
            assert annotation is result_annotations[field_name]


@patch('fastapi_toolkit.dependencies.filter.Query', QueryMock)
class TestFilterQuery:
    @pytest.fixture
    def in_mock(self) -> MagicMock:
        return MagicMock()

    @pytest.fixture
    def eq_mock(self) -> MagicMock:
        return MagicMock()

    @pytest.fixture
    def title_mock(self, in_mock: MagicMock, eq_mock: MagicMock) -> MagicMock:
        return MagicMock(
            in_=in_mock, __eq__=eq_mock, type=MagicMock(python_type=str)
        )

    def test_filter_query__ok(
            self, in_mock: MagicMock, eq_mock: MagicMock, title_mock: MagicMock
    ):
        test_value = 'test_value'
        test_fields = {'title': filter_(title_mock, (in_, eq)), }
        queries = filter_by_fields(test_fields)(
            title__in=[test_value],
            title__eq=test_value,
        ).filter_query

        assert len(queries) == 2
        in_mock.assert_called_once_with([test_value])
        eq_mock.assert_called_once_with(test_value)

    def test_filter_query__cast_type_ok(
            self, eq_mock: MagicMock, title_mock: MagicMock
    ):
        test_value = '100'
        result_value = 101
        test_fields = {
            'title': filter_(title_mock, (eq, ), lambda x: int(x) + 1)
        }
        queries = filter_by_fields(test_fields)(
            title__eq=test_value,
        ).filter_query

        assert len(queries) == 1
        eq_mock.assert_called_once_with(result_value)



import importlib
import os

from pydantic.tools import lru_cache

from fastapi_toolkit.conf.global_settings import GlobalSettings

__all__ = (
    'settings',
)

from fastapi_toolkit.logging import set_up_logging


class Settings:
    ENVIRONMENT_VARIABLE = 'SETTINGS_MODULE'

    @classmethod
    @lru_cache
    def load(cls, *args, **kwargs):
        settings_module = os.environ.get(cls.ENVIRONMENT_VARIABLE)
        if settings_module:
            module = importlib.import_module(settings_module)
            application_settings_class = getattr(module, 'Settings')
            _settings = application_settings_class(*args, **kwargs)
        else:
            _settings = GlobalSettings(*args, **kwargs)
        set_up_logging(_settings)
        return _settings


settings = Settings.load()



import logging
from contextlib import asynccontextmanager
from typing import ContextManager

import aioredis

from fastapi_toolkit.conf import settings

__all__ = (
    'get_redis_pool',
    'create_redis_pool',
    'close_redis_pool',
    'redis_connection_pool'
)

logger = logging.getLogger(settings.project)
_redis_pool: aioredis.ConnectionPool = None  # noqa


def get_redis_pool() -> aioredis.ConnectionPool:
    if _redis_pool is None:
        raise ValueError(
            'Redis pool is undefined. '
            'You need to call `createredis_pool` first.'
        )
    return _redis_pool


async def create_redis_pool() -> aioredis.ConnectionPool:
    global _redis_pool
    await close_redis_pool()
    if settings.redis_dsn:
        _redis_pool = aioredis.ConnectionPool.from_url(
            settings.redis_dsn,
            max_connections=settings.redis_max_connections
        )
    else:
        logger.warning('Can not create redis pool. Redis dsn is not defined.')
    return _redis_pool


async def close_redis_pool():
    global _redis_pool
    if _redis_pool is not None:
        await _redis_pool.disconnect()


@asynccontextmanager
async def redis_connection_pool() -> ContextManager[aioredis.Redis]:
    try:
        yield aioredis.Redis(connection_pool=get_redis_pool())
    finally:
        pass



from typing import Optional

from pydantic import BaseModel


class BookSchema(BaseModel):
    id: int
    title: str
    page_amount: int

    class Config:
        orm_mode = True


class BookSchemaCreate(BaseModel):
    title: Optional[str] = None
    page_amount: Optional[int] = None



from setuptools import (
    find_packages,
    setup,
)

FILE_NAME = 'VERSION'


with open(FILE_NAME) as file:
    version = file.read()

with open('requirements.txt') as file:
    requirements = file.readlines()

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='fastapi_toolkit',
    version=version,
    description='Package for creating web app using fastapi',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/timaakulich/fastapi_toolkit',
    author='Timofey Akulich',
    author_email='tima.akulich@gmail.com',
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'fastapi-toolkit = fastapi_toolkit.console_scripts:main'
        ]
    },
)



from dataclasses import make_dataclass
from datetime import datetime
from typing import (
    Dict,
    List,
)

from fastapi import Query

from fastapi_toolkit.filters import filter_type
from fastapi_toolkit.filters.func import (
    LIST_OPERATORS,
    OPERATORS_TYPES,
)


def filter_by_fields(available_fields: Dict[str, filter_type]):
    field_definitions = []
    for field_name, field_filter in available_fields.items():
        for operator_name, operator_func in field_filter.operators.items():
            is_list = operator_func in LIST_OPERATORS
            field_type = (
                OPERATORS_TYPES.get(operator_func)
                or field_filter.query_param_type
                or field_filter.field.type.python_type
            )
            name = f'{field_name}__{operator_name}'
            field_definitions.append(
                (
                    name,
                    List[field_type] if is_list else field_type,
                    Query(None, alias=f'{name}{[] if is_list else ""}')
                )
            )

    @property
    def filter_query(self):
        _filter_query = []
        for _field in self.__dataclass_fields__:
            if (value := getattr(self, _field, None)) is None:
                continue
            if isinstance(value, datetime):
                value = value.replace(tzinfo=None)
            _field_name, operator = _field.rsplit('__', 1)
            cast_type = available_fields[_field_name].cast_type
            _filter_query.append(
                available_fields[_field_name].operators[operator](
                    available_fields[_field_name].field,
                    cast_type(value)
                )
            )
        return _filter_query

    return make_dataclass('FilterQueryParams', field_definitions, namespace={
        'filter_query': filter_query
    })



import typer
from alembic.command import (
    downgrade,
    upgrade,
)
from alembic.config import Config

from fastapi_toolkit.conf import settings


def migrate(
        revision: str = typer.Option('head'),
        down: bool = typer.Option(False)
):
    alembic_config = Config(settings.alembic_config)
    method = downgrade if down else upgrade
    method(alembic_config, revision)



from sqlalchemy import (
    Column,
    Integer,
    String,
)

from fastapi_toolkit.db import BaseModel

__all__ = (
    'Book',
)


class Book(BaseModel):
    id = Column(Integer, primary_key=True)
    title = Column(String(length=42), nullable=False)
    page_amount = Column(Integer, nullable=False)
